"""Provides class and functions to build a model of a chrooted system
and build an sh script which will construct that system.
"""

__version__ = "$Revision: 1.14 $"
__author__ = "Peter Hendrickson <pdh@wiredyne.com>"

# License:  This code is in the public domain.  Any of this code may be
#           freely reproduced or adapted, even without mentioning the
#           source.

# Warranty: IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR
#           DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
#           DAMAGES ARISING OUT OF THE USE OF THIS CODE, EVEN IF THE
#           AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#           THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES,
#           INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#           MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
#           CODE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THERE
#           IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
#           SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

import os.path
import os
import popen2
import re
import stat

class ChrootBuilderError(Exception):
    pass

class ChrootModel:
    """A ChrootModel models the new chrooted environment.

    Files, links, directories, and devices which will be needed are
    fed in.  Additional dependencies are found by following links and
    looking for needed shared libraries.  Once everything is found, an
    sh script which creates the new chroot environment can be created.
    The model is Python independent, although it can be used to
    construct a chrooted environment which has Python available.
    """

    def __init__(self, newroot, no_suid_files=True, no_gid_files=True):
        """Initialize using newroot as directory to be given to chroot.

        newroot must not be "/" and must be an absolute path or an
        exception will be thrown.

        If no_suid_files is True, an exception is thrown if any file
        dependency has the setuid bit set.

        If no_gid_files is True, an exception is thrown if any file
        depency has the setgid bit set.
        """

        if not os.path.isabs(newroot):
            raise ChrootBuilderError("Newroot must be an absolute path: %s" \
                                     % (newroot))
        if os.path.exists(newroot) and os.path.samefile("/", newroot):
            raise ChrootBuilderError("Newroot is same as root!  Very bad! " \
                                     "(Can mess up your system.)")
        self.newroot = newroot
        self.no_suid_files = no_suid_files
        self.no_gid_files = no_gid_files
        self.hopper = []
        self.processed = {}
        self.directories = {}
        self.files = {}
        self.chardevs = {}
        self.blockdevs = {}
        self.links = {}

    def add_files(self, *files):
        """Add files (a list) to the set that will be processed."""

        for file in files:
            absfile = os.path.abspath(file)
            if not os.path.exists(absfile):
                raise ChrootBuilderError("Can't add a file which " \
                                         "doesn't exist: " \
                                         "%s (Originally %s)" % (absfile, file))
            self.hopper.append(absfile)

    def process(self):
        """Resolve all dependencies."""

        while self.hopper:
            generic_file = unshift(self.hopper)
            if not self.processed.has_key(generic_file):
                if os.path.islink(generic_file):
                    link = generic_file
                    (dir, linkname) = os.path.split(link)
                    if not self.processed.has_key(dir):
                        self.hopper.append(dir)
                    target = os.readlink(link)
                    if not os.path.isabs(target):
                        abstarget = os.path.normpath(os.path.join(dir, target))
                    else:
                        abstarget = target
                    if not os.path.exists(abstarget):
                        raise ChrootBuilderError("Link %s points to " % (link) + \
                                                 "non-existent target " \
                                                 "%s" % (abstarget))
                    self.hopper.append(abstarget)
                    self.links[link] = (dir, linkname, target)
                    self.hopper.append(dir)
                elif os.path.isfile(generic_file):
                    file = generic_file
                    file_stat = os.stat(file)
                    if self.no_suid_files and suid_mode(file_stat.st_mode):
                        raise ChrootBuilderError("%s has the setuid " % (file) + \
                                                 "bit set - not allowed.")
                    if self.no_gid_files and sgid_mode(file_stat.st_mode):
                        raise ChrootBuilderError("%s has the gid " % (file) + \
                                                 "bit set - not allowed.")
                    (dir, filename) = os.path.split(file)
                    if not self.processed.has_key(dir):
                        self.hopper.append(dir)
                    if is_ELF_file(file):
                        shared_libs = get_shared_libs(file)
                        self.hopper.extend(shared_libs)
                    self.files[file] = 1
                elif os.path.isdir(generic_file):
                    dir = generic_file
                    dirmode = directory_mode(dir)
                    (parentdir, currentdir) = os.path.split(dir)
                    if len(currentdir) != 0: # I.e., dir is not root
                        self.hopper.append(parentdir)
                    self.directories[dir] = dirmode
                elif ischaracter(generic_file):
                    chardev = generic_file
                    (dir, devname) = os.path.split(chardev)
                    if not self.processed.has_key(dir):
                        self.hopper.append(dir)
                    (major, minor) = device_numbers(chardev)
                    devmode = device_mode(chardev)
                    self.chardevs[chardev] = (major, minor, devmode)
                elif isblock(generic_file):
                    blockdev = generic_file
                    (dir, devname) = os.path.split(chardev)
                    if not self.processed.has_key(dir):
                        self.hopper.append(dir)
                    (major, minor) = device_numbers(blockdev)
                    devmode = device_mode(blockdev)
                    self.blockdevs[blockdev] = (major, minor, devmode)
                else:
                    raise ChrootBuilderError("%s is not a link or a file." \
                                             % (generic_file))
                self.processed[generic_file] = 1

    def dir_script(self):
        """Return string containing script to generate directories."""

        dirs = self.directories.keys()
        if len(dirs) == 0:
            return ''
        s = "# Create Directories\n"
        dirs.sort()
        for dir in dirs:
            newdir = os.path.join(self.newroot, dir[1:])
            s += "mkdir %s\n" % (newdir)
        s += "\n"
        s += "# Set Directory Permissions\n"
        dirs.reverse()
        for dir in dirs:
            newdir = os.path.join(self.newroot, dir[1:])
            dirmode = self.directories[dir]
            s += "chmod %s %s\n" % (oct(dirmode), newdir)
        return s

    def file_script(self):
        """Return string containing script to generate files."""

        files = self.files.keys()
        if len(files) == 0:
            return ''
        s = "# Copy Files\n"
        dirs = {}
        for file in files:
            (dir, filename) = os.path.split(file)
            if not dirs.has_key(dir):
                dirs[dir] = {}
            dirs[dir][filename] = 1
        dirlist = dirs.keys()
        dirlist.sort()
        for dir in dirlist:
            files = dirs[dir].keys()
            files.sort()
            for file in files:
                oldfile = os.path.join(dir, file)
                newdir = os.path.join(self.newroot, dir[1:])
                newfile = os.path.join(newdir, file)
                s += "cp -p %s %s\n" % (oldfile, newfile)
        return s

    def device_script(self):
        """Return string containing script to generate devices."""

        if (len(self.chardevs.keys()) == 0) and \
           (len(self.blockdevs.keys()) == 0):
            return ''
        s = "# Create Devices\n"
        for chardev in self.chardevs:
            (dir, devname) = os.path.split(chardev)
            newdir = os.path.join(self.newroot, dir[1:])
            newdev = os.path.join(newdir, devname)
            (major, minor, devmode) = self.chardevs[chardev]
            s += "mknod %s c %s %s\n" % (newdev, major, minor)
            s += "chmod %s %s\n" % (oct(devmode), newdev)
        for blockdev in self.blockdevs:
            (dir, devname) = os.path.split(blockdev)
            newdir = os.path.join(self.newroot, dir[1:])
            newdev = os.path.join(newdir, devname)
            (major, minor, devmode) = self.blockdevs[blockdev]
            s += "mknod %s b %s %s\n" % (newdev, major, minor)
            s += "chmod %s %s\n" % (newdev, oct(devmode))
        return s

    def link_script(self):
        """Return string containing script to generate links."""

        if len(self.links.keys()) == 0:
            return ''
        s = "# Create Links\n"
        s += "ORIGDIR=`pwd`\n"
        dirs = {}
        for link in self.links.keys():
            (dir, link, target) = self.links[link]
            if not dirs.has_key(dir):
                dirs[dir] = {}
            if dirs[dir].has_key(link):
                raise ChrootBuilderError("Link is apparently defined " \
                                         "twice: %s" % (link))
            dirs[dir][link] = target
        dirlist = dirs.keys()
        dirlist.sort()
        for dir in dirlist:
            newdir = os.path.join(self.newroot, dir[1:])
            s += "cd %s\n" % (newdir)
            links = dirs[dir].keys()
            links.sort()
            for link in links:
                target = dirs[dir][link]
                s += "ln -s %s %s\n" % (target, link)
        s += "cd $ORIGDIR\n"
        return s

    def script(self):
        """Return string containing complete script.

        Script will create the complete environment represented by a
        ChrootModel instance.
        """

        newscript = ""
        ds = self.dir_script()
        fs = self.file_script()
        ls = self.link_script()
        devs = self.device_script()
        newscript = ''
        for s in [ds, fs, ls, devs]:
            if len(s) > 0:
                if len(newscript) == 0:
                    newscript = s
                else:
                    newscript += "\n" + s
        return newscript

def is_ELF_file(file):
    """Return True if file is an ELF format file.

    Note: False is returned if file is a link.
    """

    if os.path.isfile(file) and not os.path.islink(file):
        fileobj = open(file, "r")
        label = fileobj.read(4)
        fileobj.close()
        if len(label) != 4:
            return False
        elif label == '\x7fELF':
            return True
        else:
            return False
    else:
        return False

getlib = re.compile(r'^\t -l\S+ => (\S+)$')

# ldd used because it's portable.
def get_shared_libs(file):
    """Return list of shared libraries used by file.

    Uses the external program "ldd" to probe ELF files.  ldd
    reports libraries needed directly or indirectly.

    If file does not exist, an exception is thrown.
    If file is not an ELF file, return an empty list.
    If file is not dynamically linked, return an empty list.
    """

    if not os.path.exists(file):
        raise ChrootBuilderError("%s does not exist." % (file))
    if not is_ELF_file(file):
        return []
    shared_libs = []
    (out, inn, err) = popen2.popen3("ldd %s" % (file))
    ldd_err = err.read()
    ldd_raw = out.read().split("\n")
    out.close()
    err.close()
    inn.close()
    if ldd_raw[0] == file + ":":
        for lddline in ldd_raw[1:-1]:
            match = getlib.search(lddline)
            if match == None:
                raise ChrootBuilderError("Unexpected line from ldd: %s" \
                                         % (lddline))
            newlib = match.groups()[0]
            shared_libs.append(newlib)
        shared_libs.sort()
        return shared_libs
    elif ldd_err == "ldd: %s: not dynamically linked\n" % (file):
        return []
    else:
        raise ChrootBuilderError("Unexpected ldd behavior. Stdout: " \
                                 "%s Stderr: %s" % (ldd_raw, ldd_err))

def suid_mode(mode):
    """Return True if mode has setuid bit set."""

    if( mode & 0004000 ):
        return True
    else:
        return False

def sgid_mode(mode):
    """Return True if mode has setgid bit set."""

    if( mode & 0002000 ):
        return True
    else:
        return False

device = re.compile(r'^[bc][rw\-]{9}\s+\d+\s+\S+\s+\S+\s+(\d+),\s+(\d+)\s+\S+\s+\d+\s+\S+\s+\S+$')

# ls used because it's portable.
def device_numbers(special):
    """Return list of major and minor device numbers for special.

    special must be a special device, otherwise an exception is thrown.
    """

    (out, inn, err) = popen2.popen3("ls -ld %s" % (special))
    ls_err = err.read()
    ls_raw = out.read()
    out.close()
    err.close()
    inn.close()
    if len(ls_err) != 0:
        raise ChrootBuilderError("Error doing ls: %s" % (ls_err))
    match = device.search(ls_raw)
    if match is None:
        raise ChrootBuilderError("Weird result from ls: %s" % (ls_raw))
    (major, minor) = match.groups()
    return [major, minor]

def device_mode(special):
    """Return mode associated with special device.

    Only the bits for user, group, and other are returned.
    Other special bits are ignored.
    """

    special_stat = os.stat(special)
    devmode = special_stat.st_mode & 0777
    return devmode

def directory_mode(directory):
    """Return mode associated with directory.

    All mode bits are returned with no masking.
    """

    dir_stat = os.stat(directory)
    dirmode = stat.S_IMODE(dir_stat.st_mode)
    return dirmode

def ischaracter(generic_file):
    """Return True if generic_file is a character device."""

    file_stat = os.stat(generic_file)
    if stat.S_ISCHR(file_stat.st_mode):
        return True
    else:
        return False

def isblock(generic_file):
    """Return True if generic_file is a block device."""

    file_stat = os.stat(generic_file)
    if stat.S_ISBLK(file_stat.st_mode):
        return True
    else:
        return False

def unshift(list):
    """Remove first element from list and return it."""

    ret = list[0]
    del list[0]
    return ret
