#!/usr/bin/env python

# usage: chroot-mixminion-netbsd.py newroot username script-name

__version__ = "$Revision: 1.1 $"
__author__ = "Peter Hendrickson <pdh@wiredyne.com>"

# License:  This code is in the public domain.  Any of this code may be
#           freely reproduced or adapted, even without mentioning the
#           source.

# Warranty: IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR
#           DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
#           DAMAGES ARISING OUT OF THE USE OF THIS CODE, EVEN IF THE
#           AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#           THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES,
#           INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#           MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
#           CODE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THERE
#           IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
#           SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

# Prints scripts which when run will make a chrooted jail for
# Mixminion.  Script should be run as root.  Specific to NetBSD, but
# could be adapted to another platform.

# An example command for mixminion for certain settings:
# env -i PATH=/bin:/usr/local/bin /usr/sbin/chroot -u mixserve -g mixserve /usr/var/chroot/mixserve /usr/pkg/bin/mixminion version

# sys built in - won't disrupt later code.
import sys

if len(sys.argv) != 4:
    print >> sys.stderr, "usage: %s newroot username script-name" % (sys.argv[0])
    sys.exit(1)

# Get all Mixminion stuff loaded.
import mixminion.Main

# Takes a while, but we need to make sure every required module gets
# loaded.
mixminion.Main.main(['/usr/pkg/bin/mixminion', 'unittests'])

# ExampleMod is left over from unittests and isn't wanted in a
# chrooted installation.
if sys.modules.has_key('ExampleMod'):
    del sys.modules['ExampleMod']

# Not loaded by unittests, but nice to have.
import mixminion.benchmark

# Take a snapshot of the dependencies.
import PythonDepends
raw_py_depends = PythonDepends.python_depends()

# With snapshot taken, we can load whatever we like.
from ChrootBuilder import *
import grp
import os
import os.path
import pwd
import stat

def main():
    newroot = sys.argv[1]
    username = sys.argv[2]
    outfilename = sys.argv[3]

    model = ChrootModel(newroot)

    py_depends = [os.path.abspath(file) for file in raw_py_depends]
    py_depends.sort()
    model.add_files(*py_depends)

    # Python complains if it doesn't have its lib-dynload directory,
    # even if nothing is in it.
    lib_dynload_dir = sys.exec_prefix + '/lib/python' + sys.version[0:3] \
                      + '/lib-dynload'
    model.add_files(lib_dynload_dir)

    # Need Python binary and shared object support library.
    model.add_files("/usr/libexec/ld.elf_so",
                    "/usr/local/bin/python")

    # Stuff needed by Mixminion.
    model.add_files("/bin/ls",
                    "/bin/rm",
                    "/usr/pkg/bin/mixminion",
                    "/usr/pkg/bin/mixminiond",
                    "/dev/random",
                    "/dev/urandom",
                    "/dev/null",
                    "/tmp")
    model.process()
    if os.path.exists(outfilename):
        raise ValueError("%s already exists.  Won't overwrite." % (outfilename))
    outfile = open(outfilename, "w")
    outfile.write("#!/bin/sh\n" \
                  "\n" \
                  "# Umask which lets other processes see but not modify.\n" \
                  "umask 0022\n" \
                  "\n")
    outfile.write(model.script() + "\n")
    outfile.write(setup_passwd_stuff(newroot, username) + "\n")
    outfile.write(mixminion_homedir(newroot, username) + "\n")
    outfile.write(lockdown(newroot))
    outfile.close()

def setup_passwd_stuff(newroot, username):
    """Set up password and group files to support username.

    NetBSD specific.  Uses the uids and gids it finds on the system.
    """

    s = "# Accounts\n"
    s += "mkdir -p %s\n" % (os.path.join(newroot, "etc"))
    root_info = pwd.getpwnam('root')
    root_desc = root_info[4]
    root_homedir = root_info[5]
    root_entry = "root:*:0:0::0:0:%s:%s:/bin/sh" % (root_desc, root_homedir)
    acct_info = pwd.getpwnam(username)
    acct_uid = acct_info[2]
    acct_gid = acct_info[3]
    acct_desc = acct_info[4]
    acct_homedir = acct_info[5]
    acct_entry = "%s:*:%s:%s::0:0:%s:%s:/sbin/nologin" % \
                 (username, acct_uid, acct_gid, acct_desc, acct_homedir)
    s += "echo '%s' > %s/etc/master.passwd\n" % (root_entry, newroot)
    s += "echo '%s' >> %s/etc/master.passwd\n" % (acct_entry, newroot)
    group_name = grp.getgrgid(acct_gid)[0]
    s += "echo 'wheel:*:0:root' > %s/etc/group\n" % (newroot)
    if group_name == username:
        s += "echo '%s:*:%s:' >> %s/etc/group\n" % (group_name, acct_gid, newroot)
    else:
        s += "echo '%s:*:%s:%s' >> %s/etc/group\n" % \
             (group_name, acct_gid, username, newroot)
    s += "pwd_mkdb -p -d %s %s/etc/master.passwd\n" % (newroot, newroot)
    return s

def mixminion_homedir(newroot, username):
    """Return script to create home directory for username and set ownership."""

    s = "# Mixminion Home Directory\n"
    acct_info = pwd.getpwnam(username)
    acct_gid = acct_info[3]
    acct_homedir = acct_info[5]
    group_name = grp.getgrgid(acct_gid)[0]
    newdir = os.path.join(newroot, acct_homedir[1:])
    s += "mkdir -p %s\n" % (newdir)
    s += "chown %s:%s %s\n" % (username, group_name, newdir)
    return s

def lockdown(newroot):
    """Return script to make most stuff unmodifiable."""

    s = "# Battening down the hatches\n"
    s += "ORIGDIR=`pwd`\n"
    s += "cd %s\n" % (newroot)
    s += "chmod -R a-w bin etc\n"
    s += "cd dev\n"
    s += "chmod a-w random urandom\n"
    s += "cd ..\n"
    s += "chmod a-w dev\n"
    s += "cd usr\n"
    s += "chmod -R a-w lib libexec local pkg\n"
    s += "chmod a-w home\n"
    s += "cd ..\n"
    s += "chmod a-w usr .\n"
    s += "chflags -R schg bin etc\n"
    s += "cd dev\n"
    s += "chflags schg random urandom\n"
    s += "cd ..\n"
    s += "chflags schg dev\n"
    s += "cd usr\n"
    s += "chflags -R schg lib libexec local pkg\n"
    s += "chflags schg home\n"
    s += "cd ..\n"
    s += "chflags schg usr .\n"
    s += "cd $ORIGDIR\n"
    return s

main()
